local _G = getfenv(0)
local ADDON_NAME, addon = ...
local LibStub = _G.LibStub
local DKDiseases = LibStub("AceAddon-3.0"):GetAddon(addon.addonName)
local L = LibStub("AceLocale-3.0"):GetLocale(addon.addonName)
local LSM = _G.LibStub:GetLibrary("LibSharedMedia-3.0")
local Masque = LibStub("Masque", true)

local SpellIds = addon.SpellIds
local SpellNames = addon.SpellNames
local SpellIcons = addon.SpellIcons
local FormatNumber = addon.FormatNumber

-- Locals for performance
local pairs = _G.pairs
local ipairs = _G.ipairs
local GetTime = _G.GetTime
local tostring = _G.tostring
local math = _G.math
local abs = _G.math.abs

local SRTracker = {}
addon:RegisterModule("SRTracker", SRTracker)
SRTracker.enabled = false

local function EventFrame_OnEvent(frame, event, ...)
	if event == "UNIT_HEALTH" then
		SRTracker:UNIT_HEALTH(event, ...)
	end
end
local EventFrames = {}
local WatchedUnits = { "target" }
for i, unit in ipairs(WatchedUnits) do
	if not EventFrames[unit] then
		local frame = _G.CreateFrame("Frame", ADDON_NAME.."_EventFrame_SRTracker_"..unit)
		frame:SetScript("OnEvent", EventFrame_OnEvent)
		EventFrames[unit] = frame
	end
end

function SRTracker:RegisterCallbacks()
	addon:RegisterCallback("TalentUpdate", "SRTracker", SRTracker.OnTalentUpdate)
	addon:RegisterCallback("StatUpdate", "SRTracker", SRTracker.OnStatUpdate)
	addon:RegisterCallback("PlayerAlive", "SRTracker", SRTracker.OnPlayerAlive)
	addon:RegisterCallback("PlayerDead", "SRTracker", SRTracker.OnPlayerDead)
	addon:RegisterCallback("TargetChanged", "SRTracker", SRTracker.OnTargetChanged)
	addon:RegisterCallback("Load", "SRTracker", SRTracker.OnLoad)
end

function SRTracker:OnInitialize()
	self:RegisterCallbacks()
end

function SRTracker:Enable()
	self:RegisterCallbacks()
end

function SRTracker:Disable()
	addon:UnregisterCallback("TalentUpdate", "SRTracker")
	addon:UnregisterCallback("StatUpdate", "SRTracker")
	addon:UnregisterCallback("PlayerAlive", "SRTracker")
	addon:UnregisterCallback("PlayerDead", "SRTracker")
	addon:UnregisterCallback("TargetChanged", "SRTracker")
	addon:UnregisterCallback("Load", "SRTracker")
end

function SRTracker:OnTalentUpdate()
	if addon.srFrame then addon.srFrame:Update() end
end

function SRTracker:OnPlayerAlive()
	if addon.srFrame and not self.db.profile.srFrame.alwaysShow then
		addon.srFrame:Hide()
	end
end

function SRTracker:OnPlayerDead()
	if addon.srFrame and not self.db.profile.srFrame.alwaysShow then
		addon.srFrame:Hide()
	end
end

function SRTracker:OnTargetChanged()
	addon.soulReaperRange = nil
	addon.targetMaxHealth = _G.UnitHealthMax("target") or 0
	local db = addon.db.profile.srFrame
	if db.enabled and not db.alwaysShow then
		SRTracker:CheckSoulReaper()
	end
end

function SRTracker:OnLoad()
	SRTracker:CreateSoulReaperCheck()
	SRTracker:UpdateSoulReaperCheck()
end

function SRTracker:CreateSoulReaperCheck()
	if addon.srFrame then return end

	local buttonName = "DKDiseases_SRCheck"
	local frame = _G.CreateFrame("Frame", "DKDiseases_SRCheck", _G.UIParent)
	addon.srFrame = frame

	frame.SetDesiredPoint = function(frame)
		frame:ClearAllPoints()
		local profile = addon.db.profile.srFrame
		local x, y = profile.x, profile.y
    	frame:SetPoint("CENTER", _G.UIParent, "CENTER", x, y)
	end
	frame:SetDesiredPoint()
	frame.SetDimensions = function(frame)
	    frame:SetWidth(addon.db.profile.srFrame.width)
	    frame:SetHeight(addon.db.profile.srFrame.height)
	end
	frame:SetDimensions()

	frame.texture = frame:CreateTexture(nil, "ARTWORK")
    frame.texture:SetAllPoints(frame)
    frame.texture:SetVertexColor(1, 1, 1)
	if SpellIcons["Soul Reaper"] then
		frame.texture:SetTexture(SpellIcons["Soul Reaper"])
	end
	frame.showing = nil

    frame.border = frame:CreateTexture(nil, "BACKGROUND")
    frame.border:SetPoint("CENTER")
    frame.border:SetWidth(frame:GetWidth())
    frame.border:SetHeight(frame:GetHeight())

	frame.value = frame:CreateFontString(nil, "OVERLAY")
    frame.value:SetPoint("TOP", frame, "BOTTOM", 0, -4)
	local ff, fh, fflags = DKDiseases:GetFontSettings()
	local font = LSM:Fetch("font", ff)
    frame.value:SetFont(font, fh, fflags)
    frame.value:SetJustifyH("CENTER")
	frame.value:SetJustifyV("CENTER")

	frame.ResetFonts = function()
		local ff, fh, fontFlags = DKDiseases:GetFontSettings()
		local font = LSM:Fetch("font", ff)
		frame.value:SetFont(font, fh, fontFlags)						
		frame.value:SetText(frame.value:GetText())
	end

	frame.Update = function(self)
		local profile = addon.db.profile.srFrame
		SRTracker:UpdateSoulReaperCheck()
		if profile.showText then
			self.value:Show()
		else
			self.value:Hide()
		end
	end

	frame.locked = true
	frame.Lock = function(self, locked)
		if locked ~= nil then
			self.locked = locked
		end
		self:EnableMouse(not self.locked)
	end
	frame.IsLocked = function(self)
		return self.locked
	end

    frame:SetMovable(true)
    frame:RegisterForDrag("LeftButton")
    frame:SetScript("OnDragStart",
        function(self, button)
			if not self.locked then
            	self:StartMoving()
			end
        end)
    frame:SetScript("OnDragStop",
        function(self)
            self:StopMovingOrSizing()
			local scale = self:GetEffectiveScale() / _G.UIParent:GetEffectiveScale()
			local x, y = self:GetCenter()
			x, y = x * scale, y * scale
			x = x - _G.GetScreenWidth()/2
			y = y - _G.GetScreenHeight()/2
			x = x / self:GetScale()
			y = y / self:GetScale()
			local profile = addon.db.profile.srFrame
			profile.x, profile.y = x, y
			profile.userPlaced = true
			self:SetUserPlaced(false)
        end)
	frame:Lock()

	if Masque then
		addon.MasqueGroup = Masque:Group("DKDiseases", SpellNames["Soul Reaper"])
		frame.MasqueButtonData = {
			--Button = frame
			Icon = frame.texture,
			FloatingBG = false,
			Cooldown = false,
			Flash = false,
			Pushed = false,
			Normal = false,
			Disabled = false,
			Checked = false,
			Border = frame.border,
			AutoCastable = false,
			Highlight = false,
			HotKey = false,
			Count = false,
			Name = false,
			Duration = false,
			AutoCast = false,
		}
		addon.MasqueGroup:AddButton(frame, frame.MasqueButtonData)
	end
end

function SRTracker:UpdateSoulReaperCheck()
	if not addon.isDK then return end

	local eventFrame = EventFrames["target"]
	if addon.db.profile.srFrame.enabled and addon.currentSpec and
		addon.db.profile.srFrame.specs[addon.currentSpec] then
		if addon.db.profile.srFrame.alwaysShow then
			if eventFrame then
				eventFrame:UnregisterEvent("UNIT_HEALTH", "target")
			end
			--self:UnregisterEvent("PLAYER_DEAD")
			--self:UnregisterEvent("PLAYER_ALIVE")
			if addon.srFrame then addon.srFrame:Show() end
		else
			if eventFrame then
				eventFrame:RegisterUnitEvent("UNIT_HEALTH", "target")
			end
			--self:RegisterEvent("PLAYER_DEAD")
			--self:RegisterEvent("PLAYER_ALIVE")
			if addon.srFrame then addon.srFrame:Hide() end
		end
	else
		if addon.srFrame then addon.srFrame:Hide() end
	end
end

function SRTracker:CalculateSoulReaper()
	if addon.currentSpec == "Frost" then
		-- Calculate use of SR vs HB
		addon.soulReaper1 = addon.avgWeaponDmg * 0.85
		addon.soulReaper2 = addon.soulReaperBase + (3.337 * addon.effectiveAP)
		addon.howlingBlast = (499 + 0.90 * addon.effectiveAP) * 
			(1 + (2.0 * addon.mastery / 100)) * addon.frostVulnerability
		local fmt = "%s - %s"
		if addon.srFrame and addon.db.profile.srFrame.enabled then
			addon.srFrame.value:SetText(fmt:format(
				FormatNumber(addon.soulReaper1 + addon.soulReaper2),
				FormatNumber(addon.howlingBlast)))
			local useSpell = "SR"
			if addon.howlingBlast * addon.spellCrit > 
				(addon.soulReaper1 * addon.critChance + 
				 addon.soulReaper2 * addon.spellCrit) then
				useSpell = "HB"
			end
			if useSpell ~= addon.srFrame.showing then
				addon.srFrame.texture:SetTexture(
					useSpell == "SR" and SpellIcons["Soul Reaper"] or 
						SpellIcons["Howling Blast"])
				addon.srFrame.showing = useSpell
			end
		end
	end
end

function SRTracker:CheckSoulReaper()
	if not addon.srFrame or not addon.db.profile.srFrame.enabled then
		if addon.srFrame then addon.srFrame:Hide() end
		return
	end
	if addon.hostile.target and addon.currentSpec == "Frost" and 
		not _G.UnitIsDeadOrGhost("player") then
		local health = _G.UnitHealth("target") or 0
		local percent = addon.targetMaxHealth > 0 and 
			health / addon.targetMaxHealth or 0
		local srRange = percent < addon.db.profile.srFrame.percent and health > 0
		if addon.db.profile.srFrame.onlyShowHB then
			if srRange and addon.srFrame.showing == "HB" then
				addon.srFrame:Show()
			else
				addon.srFrame:Hide()
			end
		else
			if srRange ~= addon.soulReaperRange then
				if srRange then
					addon.srFrame:Show()
				else
					addon.srFrame:Hide()
				end
				addon.soulReaperRange = srRange
			end
		end
	else
		addon.srFrame:Hide()
	end
end

function SRTracker:UNIT_HEALTH(event, unit)
	if unit == "target" then
		self:CheckSoulReaper()
	end
end

function SRTracker:GetOptions()
	return "SRTracker", SpellNames["Soul Reaper"], self:GetModuleOptions()
end

function SRTracker:GetModuleOptions()
	local srTrackerOpts = {
		generalOptions = {
			order = 2,
			type = "header",
			name = L["General Options"],
		},
		srEnabled = {
			name = L["Enabled"],
			desc = L["Enabled"],
			type = "toggle",
			order = 100,
			set = function(info, val)
				addon.db.profile.srFrame.enabled = val
				DKDiseases:ToggleStatTracking()
				if addon.srFrame then addon.srFrame:Update() end
			end,
			get = function(info)
				return addon.db.profile.srFrame.enabled
			end,
		},
		showValue = {
			name = L["Value"],
			desc = L["Value"],
			type = "toggle",
			order = 110,
			set = function(info, val)
				addon.db.profile.srFrame.showText = val
				if addon.srFrame then addon.srFrame:Update() end
			end,
			get = function(info)
				return addon.db.profile.srFrame.showText
			end,
		},
		srPercent = {
			order = 120,
			name = L["Percent"],
			desc = L["Percent"],
			type = "range",
			min = 0.01,
			max = 1.0,
			step = 0.01,
			bigStep = 0.05,
			isPercent = true,
			set = function(info, val)
				addon.db.profile.srFrame.percent = val
			end,
			get = function(info, val)
				return addon.db.profile.srFrame.percent
			end,
			disabled = function()
				return addon.db.profile.srFrame.alwaysShow
			end,
		},
		alwaysShow = {
			name = L["Always Show"],
			desc = L["Always Show"],
			type = "toggle",
			order = 115,
			set = function(info, val)
				addon.db.profile.srFrame.alwaysShow = val
				addon.soulReaperRange = nil
				if addon.srFrame then
					if val then addon.srFrame:Show() else addon.srFrame:Hide() end
					self:UpdateSoulReaperCheck()
				end
			end,
			get = function(info)
				return addon.db.profile.srFrame.alwaysShow
			end,
		},
		onlyShowHB = {
			name = L["Only Show HB"],
			desc = L["OnlyShowHB_Desc"],
			type = "toggle",
			order = 115,
			set = function(info, val)
			    addon.db.profile.srFrame.onlyShowHB = val
			end,
			get = function(info)
				return addon.db.profile.srFrame.onlyShowHB
			end,
		},
		width = {
			order = 120,
			name = L["Width"],
			desc = L["Width"],	
			type = "range",
			min = 20,
			max = 500,
			step = 1,
			set = function(info, val)
				addon.db.profile.srFrame.width = val
				if addon.srFrame then addon.srFrame:SetDimensions() end
			end,
			get = function(info, val)
				return addon.db.profile.srFrame.width
			end,
		},
		height = {
			order = 130,
			name = L["Height"],
			desc = L["Height"],	
			type = "range",
			min = 20,
			max = 500,
			step = 1,
			set = function(info, val)
				addon.db.profile.srFrame.height = val
				if addon.srFrame then addon.srFrame:SetDimensions() end
			end,
			get = function(info, val)
				return addon.db.profile.srFrame.height
			end,
		},
		x = {
			order = 140,
			name = L["X Offset"],
			desc = L["XOffset_Desc"],	
			type = "range",
			softMin = -math.floor(_G.GetScreenWidth()),
			softMax = math.floor(_G.GetScreenWidth()),
			bigStep = 1,
			set = function(info, val)
				addon.db.profile.srFrame.x = val
				if addon.srFrame then addon.srFrame:SetDesiredPoint() end
			end,
			get = function(info, val)
				return addon.db.profile.srFrame.x
			end,
		},
		y = {
			order = 150,
			name = L["Y Offset"],
			desc = L["YOffset_Desc"],	
			type = "range",
			softMin = -math.floor(_G.GetScreenHeight()),
			softMax = math.floor(_G.GetScreenHeight()),
			bigStep = 1,
			set = function(info, val)
				addon.db.profile.srFrame.y = val
				if addon.srFrame then addon.srFrame:SetDesiredPoint() end
			end,
			get = function(info, val)
				return addon.db.profile.srFrame.y
			end,
		},
	}
	return srTrackerOpts
end
